import re

from fake_useragent import utils


def get_browsers(verify_ssl=True):
    # https://github.com/hellysmile/fake-useragent/pull/110
    html = utils.get(utils.settings.BROWSERS_STATS_PAGE, verify_ssl=verify_ssl)
    html = html.decode("utf-8")
    html = html.split('<table class="ws-table-all notranslate">')[1]
    html = html.split("</table>")[0]

    pattern = r'\.asp">(.+?)<'
    browsers = re.findall(pattern, html, re.UNICODE)

    browsers = [utils.settings.OVERRIDES.get(browser, browser) for browser in browsers]

    pattern = r'td\sclass="right">(.+?)\s'
    browsers_statistics = re.findall(pattern, html, re.UNICODE)

    return list(zip(browsers, browsers_statistics))



import random
from dataclasses import dataclass
from typing import List, Optional, Tuple, Union
from urllib.parse import urlparse
from uuid import UUID

import faker
import inflection
import log
import us
import zipcodes
from parse import parse
from playwright.sync_api import Browser as PlaywrightBrowser
from playwright.sync_api import ElementHandle as PlaywrightElement
from splinter.browser import ChromeWebDriver, FirefoxWebDriver
from splinter.driver import ElementAPI as SplinterElements

__all__ = ["URL"]

SplinterBrowser = Union[ChromeWebDriver, FirefoxWebDriver]
GenericBrowser = Union[PlaywrightBrowser, SplinterBrowser]

PlaywrightElements = List[PlaywrightElement]
GenericElement = Union[PlaywrightElements, SplinterElements]


def is_identifier(value: str) -> bool:
    if value.isnumeric():
        return True
    if "_" in value:
        return True
    try:
        UUID(value)
    except ValueError:
        return False
    else:
        return True


class URL:

    ROOT = "@"

    def __init__(self, url_or_domain: str, path: Optional[str] = None):
        if path == self.ROOT:
            self.value = f"https://{url_or_domain}"
        elif path:
            path = ("/" + path).rstrip("/").replace("//", "/")
            self.value = f"https://{url_or_domain}" + path
        else:
            self.value = str(url_or_domain).rstrip("/")

    def __repr__(self):
        return f"URL({self.value!r})"

    def __str__(self):
        return self.value

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return str(self) == str(other).strip("/")

        if self.domain != other.domain:
            return False
        if self.path == other.path:
            return True

        result = parse(self.path, other.path)
        if not result:
            return False

        for value in result.named.values():
            if value == self.ROOT or "/" in value:
                return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __contains__(self, value):
        return value in self.value

    @property
    def domain(self) -> str:
        return urlparse(self.value).netloc

    @property
    def path(self) -> str:
        path = urlparse(self.value).path.strip("/")
        return path if path else self.ROOT

    @property
    def fragment(self) -> str:
        return urlparse(self.value).fragment.replace("/", "_").strip("_")

    def detect_patterns(self) -> Tuple["URL", bool]:
        updated = False

        parts = self.path.split("/")
        placeholder = 0
        for index, part in enumerate(parts):
            if is_identifier(part):
                placeholder += 1
                name = inflection.singularize(parts[index - 1])
                if name.isnumeric() or "{" in name:
                    name = f"id_{placeholder}"
                parts[index] = "{" + name + "}"
                updated = True

        return URL(self.domain, "/".join(parts)), updated


ALIASES = {
    "birthday": "date_of_birth",
    "cell_phone": "phone_number",
    "email_address": "email",
    "email": "email_address",
    "phone": "phone_number",
    "prefix": "honorific",
    "street_address": "address",
    "zip_code": "postcode",
    "zip": "zip_code",
}

STREETS = ["First", "Second", "Third", "Fourth", "Park", "Main"]


@dataclass
class Person:
    honorific: str
    first_name: str
    last_name: str
    date_of_birth: str
    phone_number: str
    email_address: str
    address: str
    city: str
    state: str
    state_abbr: str
    county: str
    zip_code: str

    @classmethod
    def random(cls, fake) -> "Person":
        if random.random() > 0.5:
            prefix, first_name, last_name = (
                fake.prefix_female,
                fake.first_name_female,
                fake.last_name,
            )
        else:
            prefix, first_name, last_name = (
                fake.prefix_male,
                fake.first_name_male,
                fake.last_name,
            )

        phone_number = str(random.randint(1000000000, 9999999999))
        email_address = f"{first_name}{last_name}@{fake.free_email_domain}".lower()

        while True:
            place = random.choice(zipcodes.filter_by())
            state = us.states.lookup(place["state"])
            if state and state.statehood_year:
                break

        if random.random() < 0.75:
            number = random.randint(50, 200)
            street = random.choice(STREETS)
            place["address"] = f"{number} {street} St."
        else:
            place["address"] = fake.street_address

        return cls(
            prefix,
            first_name,
            last_name,
            fake.birthday,
            phone_number,
            email_address,
            place["address"],
            place["city"],
            state.name,
            state.abbr,
            place["county"],
            place["zip_code"],
        )

    def __str__(self):
        return f"{self.name} <{self.email}>"

    def __getattr__(self, name):
        try:
            alias = ALIASES[name]
        except KeyError:
            return super().__getattribute__(name)
        else:
            log.debug(f"Mapped fake attribute {alias!r} to {name!r}")
            return getattr(self, alias)

    @property
    def name(self) -> str:
        return f"{self.first_name} {self.last_name}"


class Fake:
    def __init__(self):
        self._generator = faker.Faker()

    def __getattr__(self, name):
        try:
            method = getattr(self._generator, name)
        except AttributeError:
            try:
                alias = ALIASES[name]
            except KeyError:
                return super().__getattribute__(name)
            else:
                log.debug(f"Mapped fake attribute {alias!r} to {name!r}")
                method = getattr(self._generator, alias)
        return method()

    @property
    def person(self) -> Person:
        return Person.random(self)

    @property
    def birthday(self) -> str:
        return self._generator.date_of_birth(minimum_age=18, maximum_age=80).strftime(
            "%m/%d/%Y"
        )



from typing import Callable, List

from playwright.sync_api import Page
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.keys import Keys

from .types import GenericBrowser, PlaywrightBrowser

__all__ = ["browser", "client", "linebreak"]

browser: GenericBrowser = None
linebreak: bool = True


class _Client:
    @property
    def windows(self):
        if isinstance(browser, PlaywrightBrowser):
            return []

        return browser.windows

    @property
    def page(self) -> Page:
        # Raises an AttributeError for non-Playwright browsers
        return browser.contexts[0].pages[0]

    @property
    def url(self) -> str:
        if isinstance(browser, PlaywrightBrowser):
            return self.page.url

        return browser.url

    @property
    def title(self) -> str:
        if isinstance(browser, PlaywrightBrowser):
            return self.page.title()

        return browser.title

    @property
    def html(self) -> str:
        if isinstance(browser, PlaywrightBrowser):
            return self.page.content()

        return browser.html

    @staticmethod
    def visit(url: str) -> None:
        if isinstance(browser, PlaywrightBrowser):
            page = browser.new_page()
            page.goto(url)
        else:
            browser.visit(url)

    def type_key(self, name: str) -> Callable:
        if isinstance(browser, PlaywrightBrowser):
            key = name.capitalize()
            return lambda: self.page.keyboard.press(key)

        key = getattr(Keys, name.upper())
        return ActionChains(browser.driver).send_keys(key).perform

    def type_key_with_modifier(self, names: List[str]) -> Callable:
        if len(names) > 2:
            raise ValueError("Multiple modifier keys are not yet supported")

        if isinstance(browser, PlaywrightBrowser):
            keys = "+".join(name.capitalize() for name in names)
            return lambda: self.page.keyboard.press(keys)

        modifier = getattr(Keys, names[0].upper())
        key = getattr(Keys, names[1].upper())
        return (
            ActionChains(browser.driver)
            .key_down(modifier)
            .send_keys(key)
            .key_up(modifier)
            .perform
        )


client = _Client()



from urllib.parse import unquote

from flask import redirect, request, url_for
from flask_api import FlaskAPI

from . import models, utils

app = FlaskAPI("Pomace")


@app.route("/")
def index():
    return redirect("/sites?url=http://example.com")


@app.route("/sites")
def pomace():
    if "url" not in request.args:
        return redirect("/")

    utils.launch_browser(restore_previous_url=False)

    url = request.args["url"]
    page = models.Page.at(url)

    for action, value in request.args.items():
        if "_" in action:
            page, _updated = page.perform(action, value, _logger=app.logger)

    data = {
        "id": page.identity,
        "url": page.url,
        "title": page.title,
        "html": page.soup.prettify(),
        "text": page.text,
        "_next": unquote(url_for(".pomace", url=page.url, _external=True)),
        "_actions": dir(page),
    }

    if not app.debug:
        utils.close_browser()

    return data



from pkg_resources import DistributionNotFound as _DistributionNotFound
from pkg_resources import get_distribution as _get_distribution

from .api import *
from .config import settings

try:
    __version__ = _get_distribution("pomace").version
except _DistributionNotFound:
    __version__ = "(local)"



# pylint: disable=no-self-use

import os
from importlib import reload

import log
from cleo.application import Application
from cleo.commands.command import Command

from . import __version__, models, prompts, server, utils
from .config import settings


class BaseCommand(Command):  # pragma: no cover
    def handle(self):
        self.configure_logging()
        self.set_directory()
        self.update_settings()
        utils.launch_browser()
        utils.locate_models()
        try:
            self.run()
        except KeyboardInterrupt:
            log.debug("User cancelled loop")
        finally:
            utils.close_browser()
            prompts.linebreak()

    def configure_logging(self):
        log.reset()
        shift = 2 if self._command.name == "exec" else 1
        log.init(verbosity=self.io.verbosity + shift)
        log.silence("datafiles", allow_warning=True)
        log.silence("urllib3.connectionpool", allow_error=True)

    def set_directory(self):
        if self.option("root"):
            os.chdir(self.option("root"))

    def update_settings(self):
        if self.option("browser"):
            settings.browser.name = self.option("browser").lower()
        prompts.browser_if_unset()

        settings.browser.headless = self.option("headless")

        if self.argument("domain"):
            if "://" in self.argument("domain"):
                settings.url = self.argument("domain")
            else:
                settings.url = "http://" + self.argument("domain")
        domains = list(set(p.domain for p in models.Page.objects.all()))
        prompts.url_if_unset(domains)


class AliasCommand(BaseCommand):
    """
    Map one domain's site definitions to another

    alias
        {source : Domain to map to another}
        {target : Domain with existing models}
    """

    def handle(self):
        self.configure_logging()
        source = self.argument("source")
        target = self.argument("target")
        settings.aliases[source] = target
        self.line(
            f"<fg=white;options=bold>{source}</> mapped to "
            f"<fg=white;options=bold>{target}</>"
        )


class CleanCommand(BaseCommand):
    """
    Remove unused actions in local site definitions

    clean
        {domain? : Limit cleaning to a single domain}
        {--r|root= : Path to directory to containing models}
    """

    def handle(self):
        self.configure_logging()
        self.set_directory()
        utils.locate_models()
        if self.argument("domain"):
            pages = models.Page.objects.filter(domain=self.argument("domain"))
        else:
            pages = models.Page.objects.all()
        for page in pages:
            page.clean(force=True)


class CloneCommand(BaseCommand):
    """
    Download site definitions from a git repository

    clone
        {url : Git repository URL containing pomace models}
        {domain? : Name of sites directory for this repository}
        {--f|force : Overwrite uncommitted changes in cloned repositories}
        {--r|root= : Path to directory to containing models}
    """

    def handle(self):
        self.configure_logging()
        self.set_directory()
        utils.locate_models()
        utils.clone_models(
            self.argument("url"),
            domain=self.argument("domain"),
            force=self.option("force"),
        )


class ExecCommand(BaseCommand):  # pragma: no cover
    """
    Run a Python script

    exec
        {script : Path to a Python script}
        {--b|browser= : Browser to use for automation}
        {--d|headless : Run the specified browser in a headless mode}
        {--r|root= : Path to directory to containing models}
    """

    def update_settings(self):
        prompts.browser_if_unset()

        if self.option("browser"):
            settings.browser.name = self.option("browser").lower()

        settings.browser.headless = self.option("headless")

    def run(self):
        utils.run_script(self.argument("script"))


class RunCommand(BaseCommand):  # pragma: no cover
    """
    Start the command-line interface

    run
        {domain? : Starting domain for the automation}
        {--b|browser= : Browser to use for automation}
        {--d|headless : Run the specified browser in a headless mode}
        {--p|prompt=* : Prompt for secrets before running}
        {--r|root= : Path to directory to containing models}
    """

    def run(self):
        for name in self.option("prompt"):
            prompts.secret_if_unset(name)

        self.clear_screen()
        page = models.auto()
        self.display_url(page)

        while True:
            action = prompts.action(page)
            if action:
                page, transitioned = page.perform(action)
                if transitioned:
                    self.clear_screen()
                    self.display_url(page)
            else:
                reload(models)
                self.clear_screen()
                page = models.auto()
                self.display_url(page)

    def clear_screen(self):
        os.system("cls" if os.name == "nt" else "clear")

    def display_url(self, page):
        self.line(f"<fg=white;options=bold>{page}</>")
        prompts.linebreak(force=True)


class ShellCommand(BaseCommand):  # pragma: no cover
    """
    Launch an interactive shell

    shell
        {domain? : Starting domain for the automation}
        {--b|browser= : Browser to use for automation}
        {--d|headless : Run the specified browser in a headless mode}
        {--r|root= : Path to directory to containing models}
    """

    def run(self):
        prompts.shell()


class ServeCommand(BaseCommand):
    """
    Start the web API interaface

    serve
        {domain? : Starting domain for the automation}
        {--b|browser= : Browser to use for automation}
        {--d|headless : Run the specified browser in a headless mode}
        {--p|prompt=* : Prompt for secrets before running}
        {--r|root= : Path to directory to containing models}
        {--debug : Run the server in debug mode}
    """

    def handle(self):
        self.configure_logging()
        self.set_directory()
        self.update_settings()
        prompts.bullet = None
        utils.locate_models()
        try:
            server.app.run(debug=self.option("debug"))
        finally:
            utils.close_browser()


application = Application("pomace", __version__)
application.add(AliasCommand())
application.add(CleanCommand())
application.add(CloneCommand())
application.add(ExecCommand())
application.add(RunCommand())
application.add(ServeCommand())
application.add(ShellCommand())



from contextlib import suppress
from copy import copy
from typing import Callable, List, Optional, Tuple

import log
from bs4 import BeautifulSoup
from cached_property import cached_property
from datafiles import datafile, field, mapper
from playwright.sync_api import ElementHandle
from selenium.common.exceptions import (
    ElementNotInteractableException,
    WebDriverException,
)
from splinter.driver.webdriver import WebDriverElement
from splinter.exceptions import ElementDoesNotExist

from . import prompts, shared
from .config import settings
from .enums import Mode, Verb
from .types import URL, GenericBrowser, PlaywrightBrowser

__all__ = ["Locator", "Action", "Page", "auto"]


@datafile(order=True)
class Locator:

    mode: str = field(default="", compare=False)
    value: str = field(default="", compare=False)
    index: int = field(default=0, compare=False)
    uses: int = field(default=0, compare=True)

    def __repr__(self) -> str:
        return f"<locator {self.mode}={self.value}[{self.index}]>"

    def __bool__(self) -> bool:
        return bool(self.mode and self.value)

    def find(self) -> Optional[WebDriverElement]:
        elements = self._mode.find(self.value)
        index = self.index

        if isinstance(shared.browser, PlaywrightBrowser):
            try:
                return elements[index]
            except IndexError:
                return None

        try:
            element = elements[index]
            assert not isinstance(element, ElementHandle)
            if index == 0 and not element.visible:
                log.debug(f"{self} found invisible element: {element.outer_html}")
                index += 1
                element = elements[index]
        except ElementDoesNotExist:
            log.debug(f"{self} unable to find element")
            return None
        else:
            self.index = index
            log.debug(f"{self} found element: {element.outer_html}")
            return element

    def score(self, value: int, *, limit: int = 0) -> bool:
        previous = self.uses

        if value > 0:
            self.uses = min(limit or 9, max(1, self.uses + value))
        else:
            self.uses = max(limit or -1, self.uses + value)

        if self.uses == previous:
            return False

        result = "Increased" if self.uses > previous else "Decreased"
        log.debug(f"{result} {self} uses to {self.uses}")
        return True

    @property
    def _mode(self) -> Mode:
        return Mode(self.mode)


@datafile
class Action:

    verb: str = ""
    name: str = ""
    locators: List[Locator] = field(default_factory=lambda: [Locator()])

    @property
    def humanized(self) -> str:
        return self._verb.humanized + " " + self.name.replace("_", " ")

    @property
    def sorted_locators(self) -> List[Locator]:
        locators = [x for x in sorted(self.locators, reverse=True) if x]
        if all(locator.uses < 0 for locator in locators):
            return locators

        if len(locators) > 1 and locators[0].uses == 0 and locators[1].uses < 0:
            log.debug("Trying new locator first")
            return [locators[0]]

        return [locator for locator in locators if locator.uses >= 0]

    @property
    def locator(self) -> Locator:
        try:
            return self.sorted_locators[0]
        except IndexError:
            return Locator("id", "placeholder")

    @property
    def valid(self) -> bool:
        return self.locator.uses >= 0

    @property
    def _max_locator_uses(self) -> int:
        try:
            return max(1, self.sorted_locators[1].uses) * 2
        except IndexError:
            return max(2, self.sorted_locators[0].uses)

    @property
    def _min_locator_uses(self) -> int:
        try:
            return min(-1, self.sorted_locators[-2].uses) * 2
        except IndexError:
            return min(-2, self.sorted_locators[-1].uses)

    def __post_init__(self):
        if self.verb and self._verb != Verb.TYPE and not self.sorted_locators:
            if settings.dev:
                log.info(f"Adding placeholder locators for {self}")
                for mode, value in self._verb.get_default_locators(self.name):
                    self.locators.append(Locator(mode, value))
            else:
                log.debug("Placeholder locators are disabled")

    def __str__(self):
        return f"{self.verb}_{self.name}"

    def __bool__(self) -> bool:
        return bool(self.verb and self.name)

    def __call__(self, *args, **kwargs) -> "Page":
        page = kwargs.pop("_page", None)
        page = self._call_method(page, *args, **kwargs)
        self.datafile.save()
        page.clean()
        return page

    def _call_method(self, page, *args, **kwargs) -> "Page":
        while self._trying_locators(*args, **kwargs):
            log.error(f"No locators able to find {self.name!r}")
            if prompts.bullet:
                shared.linebreak = False
            mode, value = prompts.mode_and_value()
            if mode and value:
                self.locators.append(Locator(mode, value))
            else:
                break

        if page and self._verb.updates:
            return page

        return auto()

    def _trying_locators(self, *args, **kwargs) -> bool:
        if self._verb == Verb.TYPE:
            if "_" in self.name:
                function = shared.client.type_key_with_modifier(self.name.split("_"))
            else:
                function = shared.client.type_key(self.name)
            self._perform_action(function, *args, **kwargs)
            return False

        for locator in self.sorted_locators:
            if locator:
                log.debug(f"Using {locator} to find {self.name!r}")
                element = locator.find()
                if element:
                    function = getattr(element, self.verb)
                    if self._perform_action(function, *args, **kwargs):
                        locator.score(+1, limit=self._max_locator_uses)
                        return False
            locator.score(-1, limit=self._min_locator_uses)

        return True

    def _perform_action(self, function: Callable, *args, **kwargs) -> bool:
        previous_url = shared.client.url
        delay = kwargs.pop("delay", None)
        wait = kwargs.pop("wait", None)
        self._verb.pre_action()
        try:
            function(*args, **kwargs)
        except ElementDoesNotExist as e:
            log.warn(e)
            return False
        except ElementNotInteractableException as e:
            log.warn(e.msg)
            return False
        except WebDriverException as e:
            log.debug(e)
            return False
        else:
            self._verb.post_action(previous_url, delay, wait)
            return True

    def clean(self, page, *, force: bool = False) -> int:
        unused_locators = []
        remove_unused_locators = force

        for locator in self.locators:
            if locator.uses <= 0:
                unused_locators.append(locator)
            if locator.uses > 1:
                remove_unused_locators = True

        log.debug(f"Found {len(unused_locators)} unused locators for {self} on {page}")
        if not remove_unused_locators:
            return 0

        if unused_locators:
            log.info(f"Cleaning up locators for {self} on {page}")
            for locator in unused_locators:
                log.info(f"Removed unused {locator}")
                self.locators.remove(locator)

        return len(unused_locators)

    @property
    def _verb(self) -> Verb:
        return Verb(self.verb)


@datafile
class Locators:
    inclusions: List[Locator] = field(default_factory=lambda: [Locator()])
    exclusions: List[Locator] = field(default_factory=lambda: [Locator()])

    @property
    def sorted_inclusions(self) -> List[Locator]:
        return [x for x in sorted(self.inclusions, reverse=True) if x]

    @property
    def sorted_exclusions(self) -> List[Locator]:
        return [x for x in sorted(self.exclusions, reverse=True) if x]

    def clean(self, page, *, force: bool = False) -> int:
        unused_inclusion_locators = []
        unused_exclusion_locators = []
        remove_unused_locators = force

        for locator in self.inclusions:
            if locator.uses <= 0:
                unused_inclusion_locators.append(locator)
            if locator.uses > 1:
                remove_unused_locators = True

        for locator in self.exclusions:
            if locator.uses <= 0:
                unused_exclusion_locators.append(locator)
            if locator.uses > 1:
                remove_unused_locators = True

        count = len(unused_inclusion_locators) + len(unused_exclusion_locators)
        log.debug(f"Found {count} unused locators for {page}")
        if not remove_unused_locators:
            return 0

        if unused_inclusion_locators:
            log.info(f"Cleaning up inclusion locators for {page}")
            for locator in unused_inclusion_locators:
                log.info(f"Removed unused {locator}")
                self.inclusions.remove(locator)

        if unused_exclusion_locators:
            log.info(f"Cleaning up exclusion locators for {page}")
            for locator in unused_exclusion_locators:
                log.info(f"Removed unused {locator}")
                self.exclusions.remove(locator)

        return len(unused_inclusion_locators) + len(unused_exclusion_locators)


@datafile(
    "./sites/{self.domain}/{self.path}/{self.variant}.yml", defaults=True, manual=True
)
class Page:

    domain: str
    path: str = URL.ROOT
    variant: str = "default"

    locators: Locators = field(default_factory=Locators)
    actions: List[Action] = field(default_factory=lambda: [Action()])

    @classmethod
    def at(cls, url: str, *, variant: str = "") -> "Page":
        if shared.client.url != url:
            log.info(f"Visiting {url}")
            shared.client.visit(url)

        if shared.client.url != url:
            log.info(f"Redirected to {url}")

        kwargs = {"domain": domain(url), "path": URL(url).path}
        variant = variant or URL(url).fragment
        if variant:
            kwargs["variant"] = variant

        page = cls(**kwargs)  # type: ignore
        log.debug(f"Loaded {page.url}: {page.title} ({page.identity})")
        return page

    @property
    def url_pattern(self) -> URL:
        return URL(self.domain, self.path)

    @property
    def exact(self) -> bool:
        return "{" not in self.path

    @property
    def browser(self) -> GenericBrowser:
        return shared.browser

    @cached_property
    def url(self) -> str:  # pylint: disable=no-self-use
        return shared.client.url

    @cached_property
    def title(self) -> str:  # pylint: disable=no-self-use
        return shared.client.title

    @cached_property
    def identity(self) -> int:
        return sum(ord(c) for c in self.text)

    @cached_property
    def text(self) -> str:
        html = copy(self.soup)
        for element in html(["script", "style"]):
            element.extract()
        lines = (line.strip() for line in html.get_text().splitlines())
        chunks = (phrase.strip() for line in lines for phrase in line.split("  "))
        return "\n".join(chunk for chunk in chunks if chunk)

    @cached_property
    def html(self) -> str:  # pylint: disable=no-self-use
        return shared.client.html

    @cached_property
    def soup(self) -> BeautifulSoup:
        return BeautifulSoup(self.html, "html.parser")

    @property
    def active(self) -> bool:
        log.debug(f"Determining if {self!r} is active")

        url = URL(domain(self.url), URL(self.url).path)
        if self.url_pattern != url:
            log.debug(f"{self!r} is inactive: URL not matched")
            return False

        log.debug("Checking that all expected elements can be found")
        for locator in self.locators.sorted_inclusions:
            if locator.find():
                if locator.score(+1):
                    self.datafile.save()
            else:
                log.debug(f"{self!r} is inactive: {locator!r} found expected element")
                return False

        log.debug("Checking that no unexpected elements can be found")
        for locator in self.locators.sorted_exclusions:
            if locator.find():
                if locator.score(+1):
                    self.datafile.save()
                log.debug(f"{self!r} is inactive: {locator!r} found unexpected element")
                return False

        log.debug(f"{self!r} is active")
        return True

    def __repr__(self):
        if self.variant == "default":
            return f"Page.at('{self.url_pattern}')"
        return f"Page.at('{self.url_pattern}', variant='{self.variant}')"

    def __str__(self):
        if self.variant == "default":
            return f"{self.url_pattern}"
        return f"{self.url_pattern} ({self.variant})"

    def __dir__(self):
        names = []
        add_placeholder = True
        for action in self.actions:
            if action:
                if action.valid:
                    names.append(str(action))
            else:
                add_placeholder = False
        if add_placeholder:
            if settings.dev:
                log.info(f"Adding placeholder action for {self}")
                self.actions.append(Action())
            else:
                log.debug("Placeholder actions are disabled")
        return names

    def __getattr__(self, value: str) -> Action:
        if "_" in value:
            verb, name = value.split("_", 1)

            with suppress(FileNotFoundError):
                self.datafile.load()

            for action in self.actions:
                if action.name == name and action.verb == verb:
                    return action

            if Verb.validate(verb, name):
                action = Action(verb, name)
                setattr(
                    action,
                    "datafile",
                    mapper.create_mapper(action, root=self.datafile),
                )
                if settings.dev:
                    self.actions.append(action)
                else:
                    log.debug("Automatic actions are disabled")
                return action

        return object.__getattribute__(self, value)

    def __contains__(self, value):
        return value in self.text

    def perform(self, name: str, value: str = "", _logger=None) -> Tuple["Page", bool]:
        _logger = _logger or log
        action = getattr(self, name)
        if action.verb in {"fill", "select"}:
            if not value:
                value = settings.get_secret(action.name) or prompts.named_value(
                    action.name
                )
                settings.update_secret(action.name, value)
            _logger.info(f"{action.humanized} with {value!r}")
            page = action(value, _page=self)
        else:
            _logger.info(f"{action.humanized}")
            page = action(_page=self)
        return page, page != self

    def clean(self, *, force: bool = False) -> int:
        count = self.locators.clean(self, force=force)

        unused_actions = []
        remove_unused_actions = force

        for action in self.actions:
            if all(locator.uses <= 0 for locator in action.locators):
                unused_actions.append(action)

        log.debug(f"Found {len(unused_actions)} unused actions for {self}")
        if unused_actions and remove_unused_actions:
            log.info(f"Cleaning up actions for {self}")
            for action in unused_actions:
                log.info(f"Removed unused {action}")
                self.actions.remove(action)

        for action in self.actions:
            count += action.clean(self, force=force)

        if count or force:
            self.datafile.save()

        return count


def auto(*, detect_patterns: bool = True) -> Page:
    matching_pages = []
    found_exact_match = False

    for page in Page.objects.filter(domain=domain(shared.client.url)):
        if page.active:
            matching_pages.append(page)
            if page.exact:
                found_exact_match = True

    if found_exact_match:
        log.debug("Removing abstract pages from matches")
        matching_pages = [page for page in matching_pages if page.exact]

    if matching_pages:
        if len(matching_pages) > 1:
            for page in matching_pages:
                log.warn(f"Multiple pages matched: {page}")
                if prompts.bullet:
                    shared.linebreak = False
        return matching_pages[0]

    if detect_patterns:
        url, updated = URL(shared.client.url).detect_patterns()
        if updated:
            log.info(f"Creating new page: {url}")
            page = Page(url.domain, url.path)
            page.datafile.save()
            return auto(detect_patterns=False)

    log.info(f"Creating new page: {shared.client.url}")
    page = Page.at(shared.client.url)
    page.datafile.save()
    return page


def domain(url: str) -> str:
    value = URL(url).domain
    with suppress(KeyError):
        value = settings.aliases[value]
        log.debug(f"Mapped {url} to alias: {value}")
    return value



# pylint: disable=expression-not-assigned,unused-variable,redefined-outer-name,unused-argument

from .. import config, prompts


def describe_browser_if_unset():
    def when_ci(expect, monkeypatch):
        monkeypatch.setenv("CI", "true")
        config.settings.browser.name = ""
        prompts.browser_if_unset()
        expect(config.settings.browser.name) == "firefox"

    def when_ci_and_override(expect, monkeypatch):
        monkeypatch.setenv("CI", "true")
        monkeypatch.setenv("BROWSER", "chrome")
        config.settings.browser.name = ""
        prompts.browser_if_unset()
        expect(config.settings.browser.name) == "chrome"


def describe_url_if_unset():
    def when_ci(expect, monkeypatch, mockbrowser):
        monkeypatch.setenv("CI", "true")
        config.settings.url = ""
        prompts.url_if_unset()
        expect(config.settings.url) == "http://example.com"


def describe_secret_if_unset():
    def when_ci(expect, monkeypatch, mockbrowser):
        monkeypatch.setenv("CI", "true")
        prompts.secret_if_unset("foobar")
        expect(config.settings.get_secret("foobar")) == "<unset>"



import os
import sys
import warnings
from functools import wraps
from importlib import import_module
from typing import Optional, Tuple, no_type_check

import log
from IPython import embed

from . import browser, enums, shared
from .config import settings

try:
    import bullet
except ImportError:
    bullet = None  # https://github.com/Mckinsey666/bullet/issues/2
    warnings.warn("Interactive CLI prompts not yet supported on Windows")


RELOAD_ACTIONS = "<reload actions>"
ADD_ACTION = "<add action>"
CANCEL = "<cancel prompt>"
DEBUG = "<start shell>"


def linebreak(*, force: bool = False):
    if not shared.linebreak or force:
        print()
        shared.linebreak = True


def offset(function):
    @wraps(function)
    def wrapper(*args, **kwargs):
        linebreak()
        value = function(*args, **kwargs)
        linebreak()
        return value

    return wrapper


def noninteractive() -> bool:
    return bullet is None or not sys.stdin.isatty() or "CI" in os.environ


@offset
def browser_if_unset():
    if settings.browser.name:
        return

    if noninteractive():
        value = os.getenv("BROWSER") or browser.NAMES[0]
        settings.browser.name = value.lower()
        return

    shared.linebreak = False
    command = bullet.Bullet(
        prompt="Select a browser for automation: ",
        bullet=" ● ",
        choices=browser.NAMES,
    )
    settings.browser.name = command.launch()


@offset
def url_if_unset(domains=None):
    if settings.url:
        return

    if noninteractive():
        settings.url = "http://example.com"
        return

    shared.linebreak = False
    if domains:
        command = bullet.Bullet(
            prompt="Starting domain: ", bullet=" ● ", choices=domains
        )
    else:
        command = bullet.Input(prompt="Starting domain: ", strip=True)
    value = command.launch()
    settings.url = f"https://{value}"


@offset
def secret_if_unset(name: str):
    if settings.get_secret(name, _log=False):
        return

    if noninteractive():
        settings.set_secret(name, "<unset>")
        return

    value = named_value(name)
    settings.set_secret(name, value)


@offset
def action(page) -> Optional[str]:
    shared.linebreak = False
    choices = [RELOAD_ACTIONS] + dir(page) + [DEBUG, ADD_ACTION]
    command = bullet.Bullet(
        prompt="Select an action: ",
        bullet=" ● ",
        choices=choices,
    )
    value = command.launch()

    if value == RELOAD_ACTIONS:
        return None

    if value == DEBUG:
        shell()
        return ""

    if value == ADD_ACTION:
        verb, name = verb_and_name()
        if verb and name:
            return f"{verb}_{name}"
        return ""

    return value


@offset
def named_value(name: str) -> Optional[str]:
    if noninteractive():
        return None

    shared.linebreak = False
    prompt = "Value for " + name.replace("_", " ") + ": "
    if "pass" in name.lower():
        command = bullet.Password(prompt=prompt)
    else:
        command = bullet.Input(prompt=prompt)
    value = command.launch()
    return value


@offset
def verb_and_name() -> Tuple[str, str]:
    choices = [CANCEL] + [verb.value for verb in enums.Verb] + [DEBUG]
    command = bullet.Bullet(
        prompt="Select element verb: ",
        bullet=" ● ",
        choices=choices,
    )
    verb = command.launch()
    linebreak(force=True)

    if verb == CANCEL:
        return "", ""

    if verb == DEBUG:
        shell()
        return "", ""

    shared.linebreak = False
    command = bullet.Input("Name of element: ")
    name = command.launch().lower().replace(" ", "_")
    return verb, name


@offset
def mode_and_value() -> Tuple[str, str]:
    if noninteractive():
        return "", ""

    choices = [CANCEL] + [mode.value for mode in enums.Mode] + [DEBUG]
    command = bullet.Bullet(
        prompt="Select element locator: ",
        bullet=" ● ",
        choices=choices,
    )
    mode = command.launch()
    linebreak(force=True)

    if mode == CANCEL:
        return "", ""

    if mode == DEBUG:
        shell()
        return "", ""

    shared.linebreak = False
    command = bullet.Input("Value to match: ")
    value = command.launch()
    return mode, value


@no_type_check
def shell():
    if settings.framework == "playwright":
        log.error("Shell mode is not supported with Playwright")
        return

    try:
        get_ipython()  # type: ignore
    except NameError:
        log.debug("Launching IPython")
        linebreak()
    else:
        return

    pomace = import_module("pomace")
    globals().update(
        {name: getattr(pomace.models, name) for name in pomace.models.__all__}
    )
    globals().update(
        {name: getattr(pomace.types, name) for name in pomace.types.__all__}
    )
    auto = pomace.models.auto

    page = auto()  # pylint: disable=unused-variable
    embed(colors="neutral")



import time
from enum import Enum
from typing import Callable, Iterator, Optional, Tuple

import inflection
import log
from selenium.webdriver.common.keys import Keys

from . import shared
from .types import GenericElement, PlaywrightBrowser


class Mode(Enum):

    NAME = "name"
    ID = "id"
    TEXT = "text"
    PARTIAL_TEXT = "text (partial)"  # extended
    VALUE = "value"
    ARIA_LABEL = "aria-label"  # extended
    CSS = "css"
    TAG = "tag"
    XPATH = "xpath"

    @property
    def finder(self) -> Callable:
        if isinstance(shared.browser, PlaywrightBrowser):
            return shared.client.page.query_selector_all

        if self is self.PARTIAL_TEXT:
            return shared.browser.links.find_by_partial_text

        if self is self.ARIA_LABEL:
            return shared.browser.find_by_css

        return getattr(shared.browser, f"find_by_{self.value}")

    def find(self, value) -> GenericElement:
        if self is self.ARIA_LABEL:
            value = f'[aria-label="{value}"]'
        elif isinstance(shared.browser, PlaywrightBrowser):
            if self is not self.CSS:
                value = f"{self.value}={value}"

        return self.finder(value)


class Verb(Enum):
    CLICK = "click"
    FILL = "fill"
    SELECT = "select"
    CHOOSE = "choose"
    TYPE = "type"

    @property
    def humanized(self) -> str:
        return (self.value.capitalize() + "ing").replace("eing", "ing")

    @classmethod
    def validate(cls, value: str, name: str) -> bool:
        values = [enum.value for enum in cls]
        if value not in values:
            return False
        # TODO: name should be validated somewhere else
        if value == "type" and not all(
            hasattr(Keys, n.upper()) for n in name.split("_")
        ):
            return False
        return True

    @property
    def updates(self) -> bool:
        return self not in {self.CLICK, self.TYPE}

    def get_default_locators(self, name: str) -> Iterator[Tuple[str, str]]:
        if self is self.CLICK:
            yield Mode.TEXT.value, inflection.titleize(name)
            yield Mode.TEXT.value, inflection.humanize(name)
            yield Mode.TEXT.value, name.replace("_", " ")
            yield Mode.VALUE.value, inflection.titleize(name)
            yield Mode.VALUE.value, inflection.humanize(name)
            yield Mode.VALUE.value, name.replace("_", " ")
        elif self in {self.FILL, self.SELECT}:
            yield Mode.NAME.value, name
            yield Mode.NAME.value, inflection.dasherize(name)
            yield Mode.ID.value, name
            yield Mode.ID.value, inflection.dasherize(name)
            yield Mode.ARIA_LABEL.value, inflection.titleize(name)
            yield Mode.CSS.value, f'[placeholder="{inflection.titleize(name)}"]'
            yield Mode.ID.value, inflection.titleize(name).replace(" ", "")

    def pre_action(self):
        if self is self.CLICK:
            if isinstance(shared.browser, PlaywrightBrowser):
                log.warn(
                    "Forcing Playwright browsers to open links "
                    "in the same window is not yet supported"
                )
                return
            shared.browser.execute_script(
                """
                Array.from(document.querySelectorAll('a[target="_blank"]'))
                .forEach(link => link.removeAttribute('target'));
                """
            )

    def post_action(
        self,
        previous_url: str,
        delay: float = 0.0,
        wait: Optional[float] = None,
    ):
        if delay:
            log.info(f"Waiting {delay} seconds before continuing")
            time.sleep(delay)

        if wait is None:
            wait = 0.0 if self.updates else 5.0
        if wait:
            log.debug(f"Waiting {wait} seconds for URL to change from {previous_url}")

        elapsed = 0.0
        start = time.time()
        logged = False
        while elapsed < wait:
            time.sleep(0.1)
            elapsed = round(time.time() - start, 1)
            current_url = shared.client.url
            if current_url != previous_url:
                log.debug(f"URL changed after {elapsed} seconds to {current_url}")
                time.sleep(delay or 0.5)
                break
            if elapsed > 1.0 and not logged:
                log.info(f"Waiting {wait} seconds for the URL to change")
                logged = True



import subprocess
import sys

import log
import splinter
from fake_useragent import UserAgent, utils
from playwright.sync_api import Error as PlaywrightError
from playwright.sync_api import sync_playwright
from splinter.exceptions import DriverNotFoundError
from webdriver_manager import chrome, firefox

from . import patched
from .config import settings
from .types import GenericBrowser, PlaywrightBrowser, SplinterBrowser

__all__ = ["launch", "resize", "save_url", "save_size", "close"]


NAMES = ["Firefox", "Chrome"]

PLAYWRIGHT_BROWSERS = {"chrome": "chromium"}

WEBDRIVER_MANAGERS = {
    "chromedriver": chrome.ChromeDriverManager,
    "geckodriver": firefox.GeckoDriverManager,
}

FALLBACK_USER_AGENT = "Mozilla/5.0 Gecko/20100101 Firefox/53.0"


def launch() -> GenericBrowser:
    if not settings.browser.name:
        sys.exit("No browser specified")

    if settings.browser.name == "open":
        settings.browser.name = NAMES[0]

    settings.framework = settings.framework.lower()
    settings.browser.name = settings.browser.name.lower()

    log.info(f"Launching {settings.browser.name!r} using {settings.framework!r}")
    try:
        function = LAUNCHERS[settings.framework]
    except KeyError:
        raise ValueError(f"Unsupported framework: {settings.framework}") from None
    else:
        return function(settings.browser.name, settings.browser.headless)


def launch_playwright_browser(name: str, headless: bool) -> PlaywrightBrowser:
    name = PLAYWRIGHT_BROWSERS.get(name, name)
    playwright = sync_playwright().start()
    try:
        browser = getattr(playwright, name)
    except AttributeError:
        sys.exit(f"Unsupported browser: {name}")

    try:
        instance = browser.launch(headless=headless)
    except PlaywrightError as e:
        if "playwright install" not in str(e):
            raise e from None
        subprocess.run((sys.executable, "-m", "playwright", "install", name))
        instance = browser.launch(headless=headless)

    setattr(instance, "_playwright", playwright)
    return instance


def launch_splinter_browser(name: str, headless: bool) -> SplinterBrowser:
    utils.get_browsers = patched.get_browsers
    options = {
        "headless": headless,
        "user_agent": UserAgent(fallback=FALLBACK_USER_AGENT)[name],
        "wait_time": 1.0,
    }
    log.debug(f"Options: {options}")
    try:
        return splinter.Browser(name, **options)
    except DriverNotFoundError:
        sys.exit(f"Unsupported browser: {name}")
    except Exception as e:  # pylint: disable=broad-except
        log.debug(str(e))

        if "exited process" in str(e):
            sys.exit("Browser update prevented launch. Please try again.")

        for driver, manager in WEBDRIVER_MANAGERS.items():
            if driver in str(e).lower():
                options["executable_path"] = manager().install()
                return splinter.Browser(name, **options)

        raise e from None


LAUNCHERS = {
    "playwright": launch_playwright_browser,
    "splinter": launch_splinter_browser,
}


def resize(browser: GenericBrowser):
    if isinstance(browser, PlaywrightBrowser):
        log.warn("Resizing Playwright browsers is not yet supported")
        return

    browser.driver.set_window_size(settings.browser.width, settings.browser.height)
    browser.driver.set_window_position(0, 0)
    size = browser.driver.get_window_size()
    log.debug(f"Resized browser: {size}")


def save_url(browser: GenericBrowser):
    if isinstance(browser, PlaywrightBrowser):
        page = browser.contexts[0].pages[0]
        url = page.url
    else:
        url = browser.url

    if "://" in url and settings.url != url:
        log.debug(f"Saving last browser URL: {url}")
        settings.url = url


def save_size(browser: GenericBrowser):
    if isinstance(browser, PlaywrightBrowser):
        log.warn("Resizing Playwright browsers is not yet supported")
        return

    size = browser.driver.get_window_size()
    if size != (settings.browser.width, settings.browser.height):
        log.debug(f"Saving last browser size: {size}")
        settings.browser.width = size["width"]
        settings.browser.height = size["height"]


def close(browser: GenericBrowser):
    if isinstance(browser, PlaywrightBrowser):
        browser.close()
        getattr(browser, "_playwright").stop()
    else:
        browser.quit()
